from distutils.core import setup
setup(
  name = 'redlemur',
  packages = ['lemur', 'lemur.utils'],
  version = '0.28',
  description = 'High level visualization library',
  author = 'Nitin Kumar',
  author_email = 'nkumar14@jhu.edu',
  url = 'https://github.com/NeuroDataDesign/lemur',
  download_url = 'https://github.com/NeuroDataDesign/lemur/archive/0.28.tar.gz',
  keywords = ['visualization', 'neuroscience', 'embeddings'],
  classifiers = [],
)
